// dmnrg.h - Density-matrix NRG
// Copyright (C) 2009-2015 Rok Zitko

#ifndef _dmnrg_h_
#define _dmnrg_h_

// Routines for saving the density matrix to disk
string saverhofn(const string & prefix, int N)
{
   return P::workdir + "/" + prefix + tostring(N);
}

// Choose!
#define LINEBYLINE

#ifdef WHOLEMATRIX
 #define saveMatrix saveMatrix_wholematrix
 #define loadMatrix loadMatrix_wholematrix
 #define saveEigen saveEigen_wholematrix
 #define loadEigen loadEigen_wholematrix
#endif

#ifdef LINEBYLINE
 #define saveMatrix saveMatrix_linebyline
 #define loadMatrix loadMatrix_linebyline
 #define saveEigen saveEigen_linebyline
 #define loadEigen loadEigen_linebyline
#endif

void saveMatrix_wholematrix(boost::archive::binary_oarchive &oa, const Matrix &m)
{
   oa << m;
}

void saveEigen_wholematrix(boost::archive::binary_oarchive &oa, const Eigen &m)
{
   oa << m;
}

void saveMatrix_linebyline(boost::archive::binary_oarchive &oa, const Matrix &m)
{
   const unsigned long size1 = m.size1();
   const unsigned long size2 = m.size2();
   oa << size1 << size2;
   
   for (int i = 0; i < size1; i++) {
      ublas::vector<t_matel> vec = ublas::vector<t_matel>(size2);
      matrix_row<const Matrix> mr = matrix_row<const Matrix>(m, i);
      vec = mr;
      oa << vec;
   }
}

void saveEigen_linebyline(boost::archive::binary_oarchive &oa, const Eigen &m)
{
   oa << m.nr << m.rmax << m.nrpost;
   oa << m.value << m.value_raw << m.absenergy;
   saveMatrix_linebyline(oa, m.matrix0);
}

void loadMatrix_wholematrix(boost::archive::binary_iarchive &ia, Matrix &m)
{
   ia >> m;
}

void loadEigen_wholematrix(boost::archive::binary_iarchive &ia, Eigen &m)
{
   ia >> m;
}

void loadMatrix_linebyline(boost::archive::binary_iarchive &ia, Matrix &m)
{
   unsigned long size1, size2;
   ia >> size1 >> size2;
   
   m = Matrix(size1, size2);
   
   for (int i = 0; i < size1; i++) {
      ublas::vector<t_matel> vec;
      ia >> vec;
      matrix_row<Matrix>(m, i) = vec;
   }
}

void loadEigen_linebyline(boost::archive::binary_iarchive &ia, Eigen &m)
{
   ia >> m.nr >> m.rmax >> m.nrpost;
   ia >> m.value >> m.value_raw >> m.absenergy;
   loadMatrix_linebyline(ia, m.matrix0);
}

void saveRho_split(unsigned int N, const string & prefix, const DensMatElements &rho)
{
   const string fn = saverhofn(prefix, N);
   ofstream MATRIXF(fn.c_str(), ios::binary | ios::out);
   if (!MATRIXF)
     my_error("Can't open file %s for writing.", fn.c_str());

   boost::archive::binary_oarchive oa(MATRIXF);

   unsigned int nr = rho.size();
   oa << nr;

   unsigned int cnt = 0;
   unsigned long total = 0;
   FOREACH(const DensMatElements::value_type &i, rho) {
     const Invar inv = i.first;
     oa << inv;
     
     ostringstream fni;
     fni << fn << "-" << cnt;
     ofstream ONEMATRIX(fni.str().c_str(), ios::binary | ios::out);
     if (!ONEMATRIX)
       my_error("Can't open file %s for writing.", fni.str().c_str());
     
     boost::archive::binary_oarchive oaone(ONEMATRIX);
     saveMatrix(oaone, i.second);
      
     if (logletter('H')) {
	cout << "(" << inv << ", " << i.second.size1() << ") ";
	cout.flush();
     }
      
     // Check each time..
     if (ONEMATRIX.bad())
       my_error("Error writing %s", fni.str().c_str());
     
     ONEMATRIX.close();

     cnt++;
     total += i.second.size1();
   }
   my_assert(cnt == nr);
   if (logletter('H')) 
     cout << "[total=" << total << " nr subspaces=" << nr << "] ";

   // Check once at the end.
   if (MATRIXF.bad()) 
     my_error("Error writing %s", fn.c_str());

   MATRIXF.close();
}

void saveRho(unsigned int N, const string & prefix, const DensMatElements &rho)
{
  my_assert(P::Ninit <= N && N <= P::Nmax-1);
  if (logletter('H')) {
     cout << "Storing density matrices... ";
     cout.flush();
  }

  saveRho_split(N, prefix, rho);
   
  if (logletter('H'))
     cout << "Done." << endl;
}

int remove(string filename)
{
   return remove(filename.c_str());
}

void loadRho_split(unsigned int N, const string & prefix, DensMatElements &rho)
{
   const string fn = saverhofn(prefix, N);
   ifstream MATRIXF(fn.c_str(), ios::binary | ios::in);
   if (!MATRIXF)
     my_error("Can't open file %s for reading", fn.c_str());

   boost::archive::binary_iarchive ia(MATRIXF);

   unsigned int nr;
   ia >> nr;

   unsigned long total = 0;
   for (unsigned int cnt = 0; cnt < nr; cnt++) {
      Invar inv;
      ia >> inv;
           
      ostringstream fni;
      fni << fn << "-" << cnt;
      ifstream ONEMATRIX(fni.str().c_str(), ios::binary | ios::in);
      if (!ONEMATRIX)
	my_error("Can't open file %s for reading.", fni.str().c_str());
     
      boost::archive::binary_iarchive iaone(ONEMATRIX);
      loadMatrix(iaone, rho[inv]);
      
      if (logletter('H')) {
	cout << "(" << inv << ", " << rho[inv].size1() << ") ";
	cout.flush();
      }
      
      my_assert(rho[inv].size1() <= diag[inv].value.size());
      
      if (ONEMATRIX.bad())
	my_error("Error reading %s", fni.str().c_str());
      
      ONEMATRIX.close();

      if (P::removefiles) {
	 if (remove(fni.str()))
	   my_error("Error removing %s", fni.str().c_str());
      }
      
      total += diag[inv].value.size();
   }
   if (logletter('H')) 
     cout << "[total=" << total << " nr subspaces=" << nr << "] ";
   
   if (MATRIXF.bad()) 
     my_error("Error reading %s", fn.c_str());

   MATRIXF.close();

   if (P::removefiles) {
      if (remove(fn))
	my_error("Error removing %s", fn.c_str());
   }   
}

void loadRho(unsigned int N, const string & prefix, DensMatElements &rho)
{
   my_assert(P::Ninit <= N && N <= P::Nmax-1);
   rho.clear();
   if (logletter('H')) {
      cout << "Loading density matrices... ";
      cout.flush();
   }
   
   loadRho_split(N, prefix, rho);

   if (logletter('H')) 
      cout << "Done." << endl;
}

string unitaryfn(unsigned int N)
{
  return P::workdir + "/" + FN_UNITARY + tostring(N);
}

/* storetransformations() stores all required information (energies,
 transformation matrices, subspace labels, dimensions of 'alpha' subspaces)
 that is needed to calculate reduced density matrix in the DM-NRG
 technique. Matrices are stored on disk as binary files.

 This function is called after the diagonalisation but prior to truncation
 (at iteration N) with DiagInfo diag, i.e. with *all calculated*
 (eigenvalue, eigenvector) pairs. NOTE: if diag=dsyevr, we effectively
 perform a truncation at the moment of the partial diagonalization!! */

void store_transformations_one(int N,
			       const DiagInfo &diag)
{
  const string fn = unitaryfn(N);
  ofstream MATRIXF(fn.c_str(), ios::binary | ios::out);
  if (!MATRIXF)
    my_error("Can't open file %s for writing.", fn.c_str());

  boost::archive::binary_oarchive oa(MATRIXF);
  oa << diag;

  if (MATRIXF.bad()) 
     my_error("Error writing %s", fn.c_str());
   
  MATRIXF.close();
}

// Split it in smaller chunks (one per invariant space)
void store_transformations_split(int N,
				 const DiagInfo &diag)
{
  const string fn = unitaryfn(N);
  ofstream MATRIXF(fn.c_str(), ios::binary | ios::out);
  if (!MATRIXF)
    my_error("Can't open file %s for writing.", fn.c_str());

  boost::archive::binary_oarchive oa(MATRIXF);

  unsigned int nr = diag.size();
  oa << nr;
   
  unsigned int cnt = 0;
  unsigned long total = 0;
  LOOP_const(diag, i) {
     const Invar I = INVAR(i);
     oa << I;
     
     ostringstream fni;
     fni << fn << "-" << cnt;
     ofstream ONEMATRIX(fni.str().c_str(), ios::binary | ios::out);
     if (!ONEMATRIX)
       my_error("Can't open file %s for writing.", fni.str().c_str());
     
     if (logletter('H')) {
       cout << "(" << I << ", " << i.second.value.size() << ") ";
       cout.flush();
     }
     
     boost::archive::binary_oarchive oaone(ONEMATRIX);
     saveEigen(oaone, i.second);

     // Here check after each write.
     if (ONEMATRIX.bad())
       my_error("Error writing %s", fni.str().c_str());
     
     ONEMATRIX.close();

     cnt++;
     
     total += i.second.value.size();
  }
  my_assert(cnt == nr);
  if (logletter('H'))
     cout << "[total=" << total << " nr subspaces=" << cnt << "] ";
   
  // Here make a single check at the end, not for each write..
  if (MATRIXF.bad()) 
     my_error("Error writing %s", fn.c_str());
   
  MATRIXF.close();
}

void store_transformations(int N,
			   const DiagInfo &diag)
{
  // P::Ninit-1 corresponds to the zero-th step, when diag contains the
  // eigenvalues from the initial diagonalization and the unitary matrices
  // are all identity matrices.
  my_assert(N >= int(P::Ninit)-1 && N <= int(P::Nmax)-1);

  if (logletter('H')) {
     cout << "Storing transformation matrices... ";
     cout.flush();
  }
   
  store_transformations_split(N, diag);
   
  if (logletter('H'))
     cout << "Done." << endl;
}   

void load_transformations_split(int N, DiagInfo &diag)
{
  const string fn = unitaryfn(N);
  ifstream MATRIXF(fn.c_str(), ios::binary | ios::in);
  if (!MATRIXF)
    my_error("Can't open file %s for reading", fn.c_str());

  boost::archive::binary_iarchive ia(MATRIXF);

  unsigned int nr; // Number of subspaces
  ia >> nr;
   
  unsigned long total = 0;
  for (unsigned int cnt = 0; cnt < nr; cnt++) {
     Invar inv;
     ia >> inv;
     
     ostringstream fni;
     fni << fn << "-" << cnt;
     ifstream ONEMATRIX(fni.str().c_str(), ios::binary | ios::in);
     if (!ONEMATRIX)
       my_error("Can't open file %s for reading.", fni.str().c_str());
     
     my_assert(diag.count(inv) == 0); // added 18.8.2015
     boost::archive::binary_iarchive iaone(ONEMATRIX);
     loadEigen(iaone, diag[inv]);
     diag[inv].perform_checks(); // basic checks, added 9.11.2015
     
     if (logletter('H')) {
	cout << "(" << inv << ", " << diag[inv].value.size() << ") ";
	cout.flush();
     }
     
     if (ONEMATRIX.bad())
       my_error("Error reading %s", fni.str().c_str());
     
     ONEMATRIX.close();
     
     total += diag[inv].value.size();
  }
  if (logletter('H'))
     cout << "[total=" << total << " nr subspaces=" << nr << "] ";
   
  if (MATRIXF.bad()) 
     my_error("Error reading %s", fn.c_str());

  MATRIXF.close();
}

void load_transformations(int N, DiagInfo &diag)
{
  my_assert(N >= int(P::Ninit)-1 && N <= int(P::Nmax)-1);
  diag.clear(); // blank state, no subspaces present

  if (logletter('H')) {
     cout << "Loading transformation matrices... ";
     cout.flush();
  }
   
  load_transformations_split(N, diag);
   
  if (logletter('H'))
     cout << "Done." << endl;
}

void remove_transformation_files_split(unsigned int N)
{
   const string fn = unitaryfn(N);
   if (remove(fn))
     my_error("Error removing %s", fn.c_str());
   
   unsigned int nr = diag.size(); // todo: cleaner solution?
   for (int cnt = 0; cnt < nr; cnt++) {
      ostringstream fni;
      fni << fn << "-" << cnt;
      if (remove(fni.str()))
	my_error("Error removing %s", fni.str().c_str());
   }
}

void remove_transformation_files(unsigned int N)
{
   remove_transformation_files_split(N);
}

// Calculation of the contribution from subspace I1 of rhoN (density
// matrix at iteration N) to rhoNEW (density matrix at iteration N-1)
void cdmI(unsigned int i, // Subspace index (alpha=1,...,P::combs)
          const Invar &I1, // Quantum numbers corresponding to subspace i
          const Matrix &rhoN, // rho^N
          const Eigen &diagI1, // contains U_{I1}
          Matrix &rhoNEW, // rho^{N-1}
          unsigned int N,
          t_factor factor) // multiplicative factor that accounts for multiplicity
{
  nrglog('D', "cdmI i=" << i << " I1=" << I1 << " factor=" << factor);
   
  // Range of indexes r and r' in matrix C^{QS,N}_{r,r'}, cf. Eq. (3.55)
  // in my dissertation.
  const unsigned int dim = rhoNEW.size2();
  my_assert(rhoNEW.size1() == rhoNEW.size2()); // quadratic matrix

  // number of states taken into account in the density-matrix at
  // *current* (Nth) stage (in subspace I1)
  const unsigned int nromega = rhoN.size2();
  my_assert(rhoN.size1() == rhoN.size2()); // quadratic matrix

  // continue only if connection exists
  if (nromega == 0 || dim == 0)
     return;

  // rmax (info[I1].rmax[i]) is the range of r in U^N_I1(omega|ri), only
  // those states that we actually kept..
  const unsigned int rmax = dm[N][I1].rmax.rmax(i);
  
  // rmax can be zero in the case a subspace has been completely truncated
  if (rmax == 0)
    return;
    
  // Otherwise, rmax must equal dim
  my_assert(rmax == dim);

  // Check range of omega: do the dimensions of C^N_I1(omega omega') and
  // U^N_I1(omega|r1) match? 
  // diagI1.getnr() is the total number of states
  // obtained after diagonalisation. It can exceed nromega due to
  // truncation.
  // We do this test at this point, to ensure rmax!=0 and dim!=0
  // and nromega!=0, otherwise there is no contribution anyway.
   
  const unsigned int I1nr = diagI1.getnr();
  if (!(nromega <= I1nr)) {
     cout << "nromega=" << nromega << " I1nr=" << I1nr << " dim=" << dim << endl;
     my_error("Range mismatch in cdmI().");
  }   

  // offset gives the offset that is added to r1,rp to find the
  // elements ri in U^N_I1(omega|ri)
  const unsigned int offset = dm[N][I1].rmax.offset(i);

  const unsigned int d1 = diagI1.matrix0.size1();
  const unsigned int d2 = diagI1.matrix0.size2();
  
  if (!(nromega <= d1 && offset+dim <= d2)) {
     cout << "nromega=" << nromega << " d1=" << d1 << endl;
     cout << "offset=" << offset << " dim=" << dim << " offset+dim=" << offset+dim
       << " d2=" << d2 << endl;
     my_error("Matrix dimensions errorin cdmI().");
  }

  const matrix_range<const Matrix> U(diagI1.matrix0, range(0, nromega), 
				     range(offset, offset+dim));

  Matrix T(dim, nromega);
  
  // T <- U^dag rhoN
  gemm(CblasConjTrans, CblasNoTrans, 
       t_factor(1.0), U, rhoN, t_factor(0.0), T);
  
  // rhoNEW <- rhoNEW + factor T U
  // Note that we are *adding* to rhoNEW
  gemm(CblasNoTrans, CblasNoTrans, 
       t_factor(factor), T, U, t_factor(1.0), rhoNEW);
}

InvarVec dmnrg_subspaces(const Invar &I)
{
  InvarVec In = input_subspaces();
  for (unsigned int i = 1; i <= P::combs; i++) {
    In[i].inverse();
    In[i].combine(I);
  }
  return In;
}

// Calculation of the shell-N REDUCED DENSITY MATRICES: 
// Calculate rho at previous iteration (N-1, rhoPrev) from rho at
// the current iteration (N, rho)

void calc_densitymatrix_iterN(DiagInfo &diag,
                              DensMatElements &rho, // input
                              DensMatElements &rhoPrev, // output
                              unsigned int N)
{
   nrglog('D', "calc_densitymatrix_iterN N=" << N);
   // loop over all subspaces at *previous* iteration
   for (Subs::const_iterator i = dm[N-1].begin(); i != dm[N-1].end(); ++i) {
      const Invar I = i->first;
      const InvarVec In = dmnrg_subspaces(I);
      unsigned int dim = i->second.kept;

      nrglog('D', "dm I=" << I << " dim=" << dim);
      
      rhoPrev[I] = Matrix(dim, dim);
      rhoPrev[I].clear();

      if (!dim) 
	continue;
      
      for (unsigned int i = 1; i <= P::combs; i++) {
         const t_factor coef = double(mult(In[i]))/double(mult(I));
	 nrglog('D', "i=" << i << " In[i]=" << In[i] << " coef=" << coef);
	 const unsigned int cnt1 = rho.count(In[i]);
	 const unsigned int cnt2 = diag.count(In[i]);
	 if (cnt1 && cnt2) 
	   cdmI(i, In[i], rho[In[i]], diag[In[i]], rhoPrev[I], N, coef);
	 //my_assert(cnt2 == 1);
	 //my_assert(cnt1 == 1);
      }
   } // loop over invariant spaces
}

// Calculation of the *** FULL REDUCED DENSITY MATIRICES ***

void calc_fulldensitymatrix_iterN(DiagInfo &diag,
                              DensMatElements &rhoFDM, // input
                              DensMatElements &rhoFDMPrev, // output
                              unsigned int N)
{
   nrglog('D', "calc_fulldensitymatrix_iterN N=" << N);
   DensMatElements rhoDD;
   if (!LAST_ITERATION(N)) 
     init_rho_FDM(rhoDD, N);

   // loop over all subspaces at *previous* iteration
   for (Subs::const_iterator i = dm[N-1].begin(); i != dm[N-1].end(); ++i) {
      const Invar I = i->first;
      const InvarVec In = dmnrg_subspaces(I);
      unsigned int dim = i->second.kept;
      nrglog('D', "fdm I=" << I << " dim=" << dim);

      rhoFDMPrev[I] = Matrix(dim, dim);
      rhoFDMPrev[I].clear();

      if (!dim)
	continue;
      
      for (unsigned int i = 1; i <= P::combs; i++) {
	// DM construction for non-Abelian symmetries: must include
	// the ratio of multiplicities as a coefficient.
	const t_factor coef = double(mult(In[i]))/double(mult(I));
        // Contribution from the KK sector. (Exception: for the N-1 iteration,
        // the rhoPrev is already initialized with the DD sector of the last
        // iteration.)
	nrglog('D', "KK sector");
	cdmI(i, In[i], rhoFDM[In[i]], diag[In[i]], rhoFDMPrev[I], N, coef);
        // Contribution from the DD sector. rhoDD -> rhoFDMPrev
        if (!LAST_ITERATION(N)) {
	   nrglog('D', "DD sector");
	   cdmI(i, In[i], rhoDD[In[i]], diag[In[i]], rhoFDMPrev[I], N, coef);
	}
      }
   } // loop over invariant spaces
}

// Returns true if all the required density matrices are already
// saved on the disk.
bool already_computed(string prefix)
{
   for (unsigned int N = P::Nmax-1; N > P::Ninit; N--) {
      const string fn = saverhofn(prefix, N-1); // note the minus 1
      unsigned int nrsubs = dm[N].size();
      for (int i = 0; i <= nrsubs; i++) {
	 ostringstream fni;
	 fni << fn << "-" << i;
	 ofstream F(fni.str().c_str(), ios::binary | ios::out);
	 if (!F) {
	    cout << fn << " not found. Computing." << endl;
	    return false;
	 }
      }
   }
   return true;
}

/* calc_densitymatrix() is called prior to starting the NRG procedure for
 the second time. Here we calculate the shell-N density matrices for all
 iteration steps. */ 

void calc_densitymatrix(DensMatElements &rho)
{
   debug("Calculating density matrices.");
   if (P::resume && already_computed(FN_RHO)) {
      cout << "Not necessary: already computed!" << endl;
      return;
   }
   
   check_trace_rho(rho); // Must be 1.
   
   TIME("DM");
   for (unsigned int N = P::Nmax-1; N > P::Ninit; N--) {
      cerr << "[DM] " << N << endl;

      DiagInfo diag;
      load_transformations(N, diag);

      DensMatElements rhoPrev;
      calc_densitymatrix_iterN(diag, rho, rhoPrev, N);
      check_trace_rho(rhoPrev); // Make sure rho is normalized to 1.
      saveRho(N-1, FN_RHO, rhoPrev);
      rho.swap(rhoPrev);
   }
}

// Sum of statistical weights from site N to the end of the Wilson chain.
double sum_wn(int N)
{
   double sum = 0.0;
   for (unsigned int n = P::Nmax-1; n >= N; n--)
     sum += STAT::wn[n];
   return sum;
}

void calc_fulldensitymatrix(DensMatElements &rhoFDM)
{
   debug("Calculating full density matrices.");
   if (P::resume && already_computed(FN_RHOFDM)) {
      cout << "Not necessary: already computed!" << endl;
      return;
   }
   
   TIME("FDM");
   for (unsigned int N = P::Nmax-1; N > P::Ninit; N--) {
      cerr << "[FDM] " << N << endl;

      DiagInfo diag;
      load_transformations(N, diag);

      DensMatElements rhoFDMPrev;
      calc_fulldensitymatrix_iterN(diag, rhoFDM, rhoFDMPrev, N);
      double tr = trace(rhoFDMPrev);
      double expected = sum_wn(N);
      double diff = (tr-expected)/expected;
      if (logletter('w')) {
	 cout << "tr[rhoFDM(" << N << ")]=" << tr << " sum(wn)=" << expected << " diff=" << diff << endl;	 
      }
      my_assert(num_equal(diff, 0.0));
      saveRho(N-1, FN_RHOFDM, rhoFDMPrev);
      rhoFDM.swap(rhoFDMPrev);
   }
}

#endif // _dmnrg_h_
