// *********** Greater correlation function ***********

   // Convention: k-loops over retained states, l-loops over
   // discarded states.

// CONJ_ME because in the fermionic Green's functions, the left-hand
// operator A is an annihilation operator, while the generated matrix
// representations correspond to creation operators.
#define op1x(a,b) CONJ_ME(op1II(a,b))
#define op2x(a,b) op2II(b,a)

// For backwards compatibility: flip energy sign for fermionic
// functions. This is because we are computing <<d^dag;d>> rather
// than <<d;d^dag>>, and because G"_AB(-omega)=G"_BA(omega).
inline double ENG(double a, int sign) 
{ 
   if (sign == S_FERMIONIC)
     return -a;
   else
     return a;
}

//#define _DEBUG

void calc_generic_FDMgt(const Eigen & diagIp, const Eigen & diagI1, 
   		      unsigned int dimp, unsigned int dim1,
		      const Matrix & op1II, const Matrix & op2II,
		      double sign, t_factor spinfactor,
		      ChainSpectrum * cs,
		      const Invar &Ip, const Invar &I1)
{
   const double wnf = STAT::wnfactor[STAT::N];

   const Matrix & rhoNIp = rhoFDM[Ip]; // hand optimised out of the loops
   const Matrix & rhoNI1 = rhoFDM[I1];

   my_assert(diagI1.getnr() == diagI1.getrmax());
   my_assert(diagIp.getnr() == diagIp.getrmax());

   unsigned int ret1 = dim1;
   unsigned int retp = dimp;
   const unsigned int all1 = diagI1.getnr(); // all states
   const unsigned int allp = diagIp.getnr();
   
   if (LAST_ITERATION()) {
     // All states are considered as discarded in the last iteration
     // of the FDM. (Note the different semantics!!)
     ret1 = 0;
     retp = 0;
   }

   // p. 16 in Costi, Zlatic, PRB  81, 235127 (2010)
    
   // Numbering I1, Ip is with respect to op1(I1,Ip), since II=(I1,Ip).
      
   // Contribution 1: sizeable at high temperatures
   for (unsigned int l1 = ret1; l1 < all1; l1++) { // D
      const t_eigen El = diagI1.absenergy(l1);
	 
      for (unsigned int lp = retp; lp < allp; lp++) { // D
	 const t_eigen Elp = diagIp.absenergy(lp);
	    
	 DELTA d;
	 d.energy = ENG(Elp - El, sign);
	 d.weight = spinfactor * op1x(l1, lp) * op2x(lp, l1) * exp(-El/P::T);
	 cs->add(d.energy, wnf * d.weight);
      }
   }

   if (LAST_ITERATION()) // only term 1 contributes
      return;
  
   // Contribution 2. No contribution at very low temperatures.
   for (unsigned int l = ret1; l < all1; l++) { // D
      const t_eigen El = diagI1.absenergy(l);
	 
      for (unsigned int k = 0; k < retp; k++) { // K
	 const t_eigen Ek = diagIp.absenergy(k);
	    
	 DELTA d;
	 d.energy = ENG(Ek - El, sign);
	 d.weight = spinfactor * op1x(l, k) * op2x(k, l) * exp(-El/P::T);
	 cs->add(d.energy, wnf * d.weight);
      }
   }

   // No contributions 3 and 4.

   // Contribution 5. Always sizeable.
   if (allp && ret1) {
      const matrix_range<const Matrix> op2cut(op2II,
					      range(0, ret1),
					      range(0, allp));
      my_assert(rhoNI1.size1() == ret1 && rhoNI1.size2() == ret1);
      Matrix op2_rho(allp, ret1);
      atlas::gemm(CblasTrans, CblasNoTrans,
		  1.0, op2cut, rhoNI1, 0.0, op2_rho);
      
      for (unsigned int k = 0; k < ret1; k++) { // K
	 const t_eigen Ek = diagI1.absenergy(k);
	 
	 for (unsigned int l = retp; l < allp; l++) { // D
	    const t_eigen El = diagIp.absenergy(l);
	    
	    DELTA d;
	    d.energy = ENG(El - Ek, sign);
	    const t_weight sum1 = op2_rho(l, k);
#ifdef _DEBUG
	    weight_bucket sum2;
	    for (unsigned int kp = 0; kp < ret1; kp++) // K
	      sum2 += op2x(l, kp) * rhoNI1(kp, k); // kp and k from I1
	    my_assert(num_equal(sum1, sum2, 1e-8));
#endif
	    d.weight = spinfactor * op1x(k, l) * t_weight(sum1);
	    cs->add(d.energy, d.weight);
	 }
      }
   }
}


// ************ Lesser correlation functions ***************

void calc_generic_FDMls(const Eigen & diagIp, const Eigen & diagI1, 
		      unsigned int dimp, unsigned int dim1,
		      const Matrix & op1II, const Matrix & op2II,
		      double sign, t_factor spinfactor,
		      ChainSpectrum * cs,
		      const Invar &Ip, const Invar &I1)
{
   const double wnf = STAT::wnfactor[STAT::N];

   const Matrix & rhoNIp = rhoFDM[Ip]; // hand optimised out of the loops
   const Matrix & rhoNI1 = rhoFDM[I1];

   my_assert(diagI1.getnr() == diagI1.getrmax()); // required for CFS and FDM
   my_assert(diagIp.getnr() == diagIp.getrmax());
   
   unsigned int ret1 = dim1;
   unsigned int retp = dimp;
   const unsigned int all1 = diagI1.getnr(); // all states
   const unsigned int allp = diagIp.getnr();
   
   if (LAST_ITERATION()) {
     // All states are considered as discarded in the last iteration
     // of the FDM. (Note the different semantics!!)
     ret1 = 0;
     retp = 0;
   }

   // p. 16 in Costi, Zlatic, PRB  81, 235127 (2010)
   
   // Contribution 1: sizeable at high temperatures
   for (unsigned int l1 = ret1; l1 < all1; l1++) { // D
      const t_eigen El = diagI1.absenergy(l1);
	 
      for (unsigned int lp = retp; lp < allp; lp++) { // D
	 const t_eigen Elp = diagIp.absenergy(lp);
	    
	 DELTA d;
	 d.energy = ENG(Elp - El, sign);
	 d.weight = spinfactor * op1x(l1, lp) * op2x(lp, l1) * (-sign) * exp(-Elp/P::T);
	 cs->add(d.energy, wnf * d.weight);
      }
   }

   if (LAST_ITERATION())
     return;
   
   // No contribution 2.
   
   // Contribution 3. No contribution at very low temperatures.
   for (unsigned int k = 0; k < ret1; k++) { // K
      const t_eigen Ek = diagI1.absenergy(k);
	 
      for (unsigned int l = retp; l < allp; l++) { // D
	 const t_eigen El = diagIp.absenergy(l);
	    
	 DELTA d;
	 d.energy = ENG(El - Ek, sign);
	 d.weight = spinfactor * op1x(k, l) * op2x(l, k) * (-sign) * exp(-El/P::T);
	 cs->add(d.energy, wnf * d.weight);
      }
   }
                      
   // Contribution 4, always sizeable.
   if (all1 && retp) {
      const matrix_range<const Matrix> op2cut(op2II,
					      range(0, all1),
					      range(0, retp));
      my_assert(rhoNIp.size1() == retp && rhoNIp.size2() == retp);
      Matrix op2_rho(all1, retp);
      atlas::gemm(CblasNoTrans, CblasNoTrans,
		  1.0, op2cut, rhoNIp, 0.0, op2_rho);
      
      for (unsigned int l = ret1; l < all1; l++) { // D
	 const t_eigen El = diagI1.absenergy(l);

	 for (unsigned int k = 0; k < retp; k++) { // K
	    const t_eigen Ek = diagIp.absenergy(k);
      
	    DELTA d;
	    d.energy = ENG(Ek - El, sign);
	    const t_weight sum1 = op2_rho(l, k);
#ifdef _DEBUG
	    weight_bucket sum2;
	    for (unsigned int kp = 0; kp < retp; kp++) // K
	      sum2 += op2x(kp, l) * rhoNIp(kp, k); // k and kp from Ip, rho symmetric!
	    my_assert(num_equal(sum1, sum2, 1e-8));
#endif
	    d.weight = (-sign) * op1x(l, k) * spinfactor * t_weight(sum1);
	    cs->add(d.energy, d.weight);
	 }
      }
   }
   
   // No contribution 5.
}
