// spectral.h - Code for handling spectral function data
// Copyright (C) 2005-2015 Rok Zitko

#ifndef _spectral_h_
#define _spectral_h_

#include <vector>
#include <utility>
#include <iostream>

// This structure is used in nrg.cc
struct DELTA 
{
     double energy;
     t_weight weight;
};

// Discrete representation of a function

typedef pair<double, double> XYPOINT;
typedef std::vector<XYPOINT> XYFUNC;

// Modified version for holding spectral information represented by
// delta peaks. "Weight" is of type t_weight (it may be complex, for example).

typedef pair<double, t_weight> DENSPOINT;
typedef std::vector<DENSPOINT> DENSFUNC;

// Save a pair of values. If imagpart=true, the imaginary part (if it
// exists, i.e. for calculations with complex numbers) is also saved.
void check_output_pair(ostream &F, double x, t_weight y, bool imagpart = false)
{
   F << x << " " << assert_isfinite(real_part(y));

   if (imagpart) 
     F << " " << assert_isfinite(imag_part(y));

   F << endl;
}

// TO DO: add possibility of saving the results as binary files!

void save_densfunc(ostream &F, const DENSFUNC &xy, bool imagpart = false)
{
   F << setprecision(P::prec_xy);
   
   for (DENSFUNC::const_iterator i = xy.begin(); i != xy.end(); i++) 
     check_output_pair(F, i->first, i->second, imagpart);
}

// STL vector of pairs. I chose this container so that I can use
// generic algorithms, for example sort()! Sorting will be performed
// according to the first number in the pair, which is exactly what we
// want to achieve.
// TO DO: this is the same as DENSPOINT. Do we really need both?? Cleanup!
typedef pair<double, t_weight> t_delta_peak;
typedef std::vector<t_delta_peak> Spikes;

inline double ENERGY(const Spikes::value_type &i) { return i.first; }
inline t_weight WEIGHT(const Spikes::value_type &i) { return i.second; }

#ifndef M_SQRTPI
#define M_SQRTPI 1.7724538509055160273
#endif

// Broadening functions

// Optimization note: prefer multiplications to divisions!!

inline double BR_LOG_GAUSS_factor(double ept, double param) 
{
  return exp(-sqr(param)*0.25) / (param*ept*M_SQRTPI); 
}

inline double BR_LOG_GAUSS_exp(double loge, double logept)
{
  return exp( -sqr(loge-logept) );
}

inline double BR_LORENTZIAN(double eta, double e1, double e2)
{
   const double diff = e1-e2;
   return (eta/M_PI)/(diff*diff+eta*eta);
}

#define SQRT2PI 2.506628274631000502416
#define INVSQRT2PI 0.3989422804014326779399

// Normalized to 1.
inline double BR_GAUSSIAN_factor(double invsigma)
{
  return INVSQRT2PI * invsigma;
}

// Optimization: Truncate results less than exp(-APPROX_0_LIMIT^2/2)=2e-22
// (for 10). We do this because exp() is an expensive operation.
#define APPROX_0_LIMIT 10.0

inline double BR_GAUSSIAN_exp(double e, double ept, double invsigma)
{
  const double scaled_diff = (e-ept) * invsigma;
  if (abs(scaled_diff) > APPROX_0_LIMIT)
     return 0.0;
   else
     return exp( -sqr(scaled_diff) * 0.5 );
}

// cf. A. Weichselbaum and J. von Delft, cond-mat/0607497

// Globally defined smoothing parameters.
double alpha;
double omega0;

// When gamma=alpha/4, the log-Gaussian broadening kernel is
// symmetric with respect to its arguments (omega,omega' or
// e,ept as called here).
#define gamma (alpha/4.0)

// Modified log-Gaussian broadening kernel. For gamma=alpha/4, the
// kernel is symmetric in both arguments.
inline double BR_L(double e, double ept)
{
  if ((e < 0.0 && ept > 0.0) || (e > 0.0 && ept < 0.0))
     return 0.0;

  if (ept == 0.0) 
    return 0.0;

  return  exp( -sqr( log(e/ept)/alpha - gamma) ) / (alpha * abs(e) * M_SQRTPI);
}

// Normalized to 1, width omega0. The kernel is symmetric in both
// arguments.
inline double BR_G(double e, double ept)
{
  return exp( -sqr( (e-ept)/omega0 ) ) / (omega0 * M_SQRTPI);
}

// Note: 'ept' is the energy of the delta peak in the raw spectrum,
// 'e' is the energy of the data point in the broadened spectrum.
inline double BR_NEW(double e, double ept)
{
  double part_l = BR_L(e, ept);
  
  // Most of the time we only need to compute part_l (loggaussian)
  if (abs(e) > omega0)
     return part_l;
   
   // Note: this is DIFFERENT from the broadening kernel proposed by
   // by Weichselbaum et al. This BR_h is a function of 'e', not
   // of 'ept'! This breaks the normalization at finite temperatures.
   // On the other hand, it gives nicer spectra when used in conjunction 
   // with the self-energy trick.
  double BR_h = exp( -sqr( log(abs(e)/omega0)/alpha ) );
  my_assert(BR_h >= 0.0 && BR_h <= 1.0);
  return part_l * BR_h + BR_G(e, ept) * (1.0-BR_h);
}

inline double BR_WVD(double e, double ept)
{
  double part_l = BR_L(e, ept);
  
  // Most of the time we only need to compute part_l (loggaussian)
  if (abs(ept) > omega0)
     return part_l;
   
  double BR_h = exp( -sqr( log(abs(ept)/omega0)/alpha ) );
  my_assert(BR_h >= 0.0 && BR_h <= 1.0);
  return part_l * BR_h + BR_G(e, ept) * (1.0-BR_h);
}

double brcut; // GLOBAL!

// Specialized broadening kernel for problems with a superconducting gap.
inline double BR_NEWSC(double e, double ept)
{
  // ept is the energy of the peak/bin in the spectrum
  if (abs(ept) > brcut) { // log-Gaussian kernel (width controlled by alpha)
    return BR_L(e, ept);
  } else { // Gaussian (width controlled by omega0)
    return BR_G(e, ept);
  }
}

CONSTFNC t_weight sum_weights(const Spikes &s)
{
  t_weight sum = 0.0;
  FOREACH(const Spikes::value_type &i, s) 
    sum += WEIGHT(i);
  return assert_isfinite(sum);
}

// Calculate "moment"-th spectral moment.
CONSTFNC t_weight moment(const Spikes &sneg, const Spikes &spos, int moment)
{
   t_weight sumA = 0.0;
   FOREACH(const Spikes::value_type &i, spos)
     sumA += WEIGHT(i) * pow(ENERGY(i), moment);
   
   t_weight sumB = 0.0;
   FOREACH(const Spikes::value_type &i, sneg)
     sumB += WEIGHT(i) * pow(-ENERGY(i), moment);
   
   t_weight sum = sumA + sumB;
   
   return assert_isfinite(sum);
}

CONSTFNC double fermi_fnc(const double omega, const double T)
{
   return 1/(1+exp(-omega/T));
}

CONSTFNC double bose_fnc(const double omega, const double T)
{
   return 1/(1-exp(-omega/T));
}

// Integrated spectral function with a kernel as in FDT for fermions
CONSTFNC t_weight fd_fermi(const Spikes &sneg, const Spikes &spos, double const T)
{
   t_weight sumA = 0.0;
   FOREACH(const Spikes::value_type &i, spos)
     sumA += WEIGHT(i) * fermi_fnc(ENERGY(i), T);
   
   t_weight sumB = 0.0;
   FOREACH(const Spikes::value_type &i, sneg)
     sumB += WEIGHT(i) * fermi_fnc(-ENERGY(i), T);

   t_weight sum = sumA + sumB;
   
   return assert_isfinite(sum);
}

// Ditto for bosons
CONSTFNC t_weight fd_bose(const Spikes &sneg, const Spikes &spos, double const T)
{
   t_weight sumA = 0.0;
   FOREACH(const Spikes::value_type &i, spos)
     sumA += WEIGHT(i) * bose_fnc(ENERGY(i), T);
   
   t_weight sumB = 0.0;
   FOREACH(const Spikes::value_type &i, sneg)
     sumB += WEIGHT(i) * bose_fnc(-ENERGY(i), T);

   t_weight sum = sumA + sumB;
   
   return sum; // removed assertion due to false triggers
}

#endif // _spectral_h_
