// Calculation of NRG chain coefficients for 2x2 matrix structure
// 
// Rok Zitko, rok.zitko@ijs.si, 2009, 2010, 2012
// $Id: nrgchain2x2.cc,v 1.3 2012/09/27 09:06:32 rokzitko Exp rokzitko $

/*
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 * 
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 * 
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * */

// CHANGE LOG
// ----------
// 18.2.2011 - supports Nambu strucure [by Oliver Bodensiek]
// 20.9.2012 - new version of the code for general 2x2 matrix structure
//             (by RZ, based on SC code by Oliver Bodensiek)
// 25.9.2012 - algorithmic refinements for better numerical stability
//           - additional checks for the correctness of input files
// 26.9.2012 - corrections in the orthonormalization procedure;
//             now it's numerically fully stable even with out-of-diagonal
//             hybridization matrix elements
//           - code cleanup
// 19.11.2014 - non-POD type arrays fix -> std::vectors

#include <iostream>
#include <fstream>
#include <iomanip>
#include <cmath>
#include <cstdlib>
#include <cassert>
#include <cfloat>
#include <utility>
#include <vector>
#include <string>
#include <map>
#include <algorithm>
#include <ctime>
#include <limits>

#include <boost/numeric/ublas/matrix.hpp>
#include <boost/numeric/ublas/io.hpp>

#include <boost/algorithm/string.hpp>  

#include <gmp.h>
#include <gmpxx.h>

#include <unistd.h>

using namespace std;
namespace ublas = boost::numeric::ublas;

typedef pair<double, double> Pair;
typedef vector<Pair> Vec;

#include "lambda.h"
#include "io.h"
#include "parser.h"
#include "load.h"

string param_fn = "param"; // file with input parameters
LAMBDA Lambda; // discretization parameter
double z; // twist parameter
unsigned int mMAX; // the number of coefficients computed (max index)
unsigned int Nmax; // the length of the Wilson chain (max index)
string discretization; // Y or C
bool rescale = false; // rescale xi or not [default changed on 10.11.2017] [Hard coded]

unsigned int preccpp; // precision for GMP

typedef vector<double> Table;

// Input to the tridiagonalisation.
Table de11_pos, de11_neg, du11_pos, du11_neg;
Table de12_pos, de12_neg, du12_pos, du12_neg;
Table de21_pos, de21_neg, du21_pos, du21_neg;
Table de22_pos, de22_neg, du22_pos, du22_neg;

void about(ostream &F = cout)
{
   F << "# Calculation of NRG coefficients [2x2 matrix structure]" << endl;
   F << "# Rok Zitko, rok.zitko@ijs.si, March 2009, Sep 2012" << endl;
   F << "# Based on work by Oliver Bodensiek, 2010" << endl;
   F << "# $Id: nrgchain2x2.cc,v 1.3 2012/09/27 09:06:32 rokzitko Exp rokzitko $" << endl;
}

void set_parameters()
{
  cout << setprecision(PREC);

  Lambda = LAMBDA(P("Lambda", 2.0));
  assert(Lambda > 1.0);

  z = P("z", 1.0);
  assert(0 < z && z <= 1.0); 

  Nmax = Pint("Nmax", 0); // Maximal site index in the Wilson chain

  mMAX = Pint("mMAX", 2*Nmax); // Maximal index of coefficients (e,f)
  assert(mMAX > 0);

  preccpp = Pint("preccpp", 2000); // Precision for GMP
  assert(preccpp > 10);

  discretization = Pstr("discretization", "y"); // Discretization scheme
  boost::algorithm::to_lower(discretization);
   
  cout << "# Lambda=" << Lambda;
  cout << " z=" << z << endl;
  cout << "# mMAX=" << mMAX;
  cout << " Nmax=" << Nmax << endl;
}

// The factor that multiplies eigenvalues of the Wilson chain Hamiltonian
// in order to obtain the eigenvalues of the true Hamiltonian (at scale D).
double SCALE(int N)
{
   if (!rescale)
      return 1.0;
   
   if (discretization == "c") 
     return (1.0 - 1./Lambda)/log(Lambda) * 
     pow(Lambda, -(N-1.0)/2.0 + 1.0 - z);
   
   if (discretization == "y")
     return (1.0+1.0/Lambda)/2.0 * 
     pow(Lambda, -(N-1.0)/2.0 + 1.0 - z);

   cerr << "Unknown discretization type " << discretization << endl;
   exit(1);
}

inline double sqr(double x) { return x*x; }

void load_tables()
{
   load("de11_pos.dat", de11_pos);
   load("de11_neg.dat", de11_neg);
   load("du11_pos.dat", du11_pos);
   load("du11_neg.dat", du11_neg);

   load("de12_pos.dat", de12_pos);
   load("de12_neg.dat", de12_neg);
   load("du12_pos.dat", du12_pos);
   load("du12_neg.dat", du12_neg);

   load("de21_pos.dat", de21_pos);
   load("de21_neg.dat", de21_neg);
   load("du21_pos.dat", du21_pos);
   load("du21_neg.dat", du21_neg);

   load("de22_pos.dat", de22_pos);
   load("de22_neg.dat", de22_neg);
   load("du22_pos.dat", du22_pos);
   load("du22_neg.dat", du22_neg);
   
   assert(de11_pos.size() == de11_neg.size());
   assert(de12_pos.size() == de12_neg.size());
   assert(de21_pos.size() == de21_neg.size());
   assert(de22_pos.size() == de22_neg.size());

   assert(du11_pos.size() == du11_neg.size());
   assert(du12_pos.size() == du12_neg.size());
   assert(du21_pos.size() == du21_neg.size());
   assert(du22_pos.size() == du22_neg.size());

   assert(de11_pos.size() == de12_pos.size());
   assert(de11_pos.size() == de21_pos.size());
   assert(de11_pos.size() == de22_pos.size());

   assert(du11_pos.size() == du12_pos.size());
   assert(du11_pos.size() == du21_pos.size());
   assert(du11_pos.size() == du22_pos.size());

   assert(du11_pos.size() == de11_pos.size());
   assert(du11_pos.size() == mMAX + 1);
}

// set up internal matrix (Nambu) space
const unsigned int dimnambu = 2;
typedef ublas::bounded_matrix<mpf_class, dimnambu, dimnambu> Matrix;

ublas::zero_matrix<mpf_class> zeroMAT(dimnambu);
ublas::identity_matrix<mpf_class> oneMAT(dimnambu);

// Arbitrary-precision trace
mpf_class spur(const Matrix &m)
{
   return m(0,0) + m(1,1);
}

#define LIMIT 1e-12

bool is1111(const Matrix &m) 
{
   mpf_class norm2 = 0.0;
   
   for (unsigned int i = 0; i < dimnambu; i++)
     for (unsigned int j = 0; j < dimnambu; j++)
       norm2 += m(i,j)*m(i,j);
   
   mpf_class norm = sqrt(norm2);
   
   // Check for equality taking into account the overall scale of mat. elements
   mpf_class diff1 = (m(0,0)-m(0,1))/norm;
   if (diff1.get_d() > LIMIT)
     return false;
   
   mpf_class diff2 = (m(0,0)-m(1,0))/norm; 
   if (diff2.get_d() > LIMIT)
     return false;

   mpf_class diff3 = (m(0,0)-m(1,1))/norm; 
   if (diff3.get_d() > LIMIT)
     return false;

   return true;
}

mpf_class matdet(const Matrix &m)
{
   return m(0,0)*m(1,1) - m(0,1)*m(1,0);
}

mpf_class matnorm2(const Matrix &m)
{
   return m(0,0)*m(0,0) + m(0,1)*m(0,1) + m(1,0)*m(1,0) * m(1,1)*m(1,1);
}

mpf_class matoutofdiag2(const Matrix &m)
{
   return m(0,1)*m(0,1) + m(1,0)*m(1,0);
}

#define DIAG_LIMIT 1e-99 // off!
Matrix matsqrt(const Matrix &m)
{
   Matrix result;
   
   mpf_class trace = spur(m);
   mpf_class det = matdet(m);
   mpf_class norm2 = matnorm2(m);
   
   cout << "matsqrt(): trace=" << trace.get_d() << " det=" << det.get_d() << " norm2=" << norm2.get_d() << endl;

   mpf_class outofdiag2 = matoutofdiag2(m);
   if (sqrt(outofdiag2.get_d()/norm2.get_d()) < DIAG_LIMIT) {
       cout << "matsqrt(): diagonal matrix detected" << endl;
      
      result = zeroMAT;
      result(0,0) = sqrt(m(0,0));
      result(1,1) = sqrt(m(1,1));
   } else {
      mpf_class s = sqrt(det);
      mpf_class t = sqrt(trace+2*s);
      
      cout << "s=" << s.get_d() << " t=" << t.get_d() << endl;
      
      mpf_class t2 = t*t; // we need to compare squares
      if (abs(t2.get_d()/norm2.get_d())<1e-12) {
	 cerr << "WARNING: t is small in matsqrt()!" << endl;
      }
      
      result = zeroMAT;
      result(0,0) = (m(0,0)+s)/t;
      result(0,1) = m(0,1)/t;
      result(1,0) = m(1,0)/t;
      result(1,1) = (m(1,1)+s)/t;
   }
   
   return result;
}

Matrix matinv(const Matrix &m)
{
   Matrix inv = zeroMAT;
       
   mpf_class outofdiag2 = matoutofdiag2(m);
   mpf_class norm2 = matnorm2(m);

   if (sqrt(outofdiag2.get_d()/norm2.get_d()) < DIAG_LIMIT) {
      cout << "matinv(): diagonal matrix detected" << endl;

      // Simple inversion
      inv(0,0) = 1/m(0,0);
      inv(1,1) = 1/m(1,1);
   } else {
      // Standard algorithm
      mpf_class d = matdet(m);
      inv(0,0) = m(1,1)/d;
      inv(0,1) = -m(1,0)/d;
      inv(1,0) = -m(0,1)/d;
      inv(1,1) = m(0,0)/d;
   }
   
   return inv;
}

void show(const string &label, const Matrix &m)
{
   cout << label << " {{";
   cout << m(0,0).get_d() << ", " << m(0,1).get_d() << "}, {";
   cout << m(1,0).get_d() << ", " << m(1,1).get_d() << "}}" << endl;
}

const bool do_fix_norm = true;
const bool check_orthogonality = false; // set "true" for debugging!

typedef std::vector<Matrix> vMatrix;
typedef ublas::matrix<Matrix> mMatrix;

Matrix calc_overlap(vMatrix &up, vMatrix &um)
{
   Matrix overlap;
   
   for (unsigned int i = 0; i < dimnambu; i++) {
      for (unsigned int j = 0; j < dimnambu; j++) {
	 mpf_class sum = 0.0;
	 for (unsigned int m = 0; m <= mMAX; m++) {
	    // alpha runs over nambu indices
	    // normalization works for dimnambu = 1,2; dimnambu = 4 to be tested
	    mpf_class tempsq = 0.0;
	    for (unsigned int alpha = 0; alpha < dimnambu; alpha++) {
	       // Note: this is effectively a sum over (m,alpha,sign)
	       // multi-index, where (sign=+,-) indicates particle or hole
	       // states.
	       tempsq += up[m](i,alpha) * up[m](j,alpha);
	       tempsq += um[m](i,alpha) * um[m](j,alpha);
	    }
	    sum += tempsq;
	 }
	 
	 overlap(i,j) = sum;
      }
   }
   
   return overlap;
}

// Fix normalization of u_{n,m}, v_{n,m} to 1. IMPORTANT: pass by
// reference!
void fix_norm(vMatrix &up, vMatrix &um)
{
  if (!do_fix_norm) {
     cerr << "Normalization disabled" << endl;
     return;
  }
   
   Matrix overlap = calc_overlap(up, um);
      
   mpf_class det = matdet(overlap);
   mpf_class tr = spur(overlap);
   mpf_class norm2 = matnorm2(overlap);
   show("overlap", overlap);
   cout << "det=" << det.get_d() << " tr=" << tr.get_d() << " norm2=" << norm2.get_d() << endl;
   
   {
      Matrix diff = overlap - oneMAT;
      mpf_class residual = sqrt(matnorm2(diff));
      cout << "Orthogonalization error=" << residual.get_d() << endl;
   }
   
   Matrix correction = matinv(matsqrt(overlap));
   
   for (unsigned int m = 0; m <= mMAX; m++) {
      up[m] = prod(correction, up[m]);
      um[m] = prod(correction, um[m]);
   }
   
   overlap = calc_overlap(up, um);

   show("overlap (after)", overlap);
   
   {
      Matrix diff = overlap - oneMAT;
      mpf_class residual = sqrt(matnorm2(diff));
      cout << "Residual orthogonalization error=" << residual.get_d() << endl;
   }
}

double chop(double val)
{
   return (abs(val) > 1e-20 ? val : 0.0);
}

#define HIGHPREC(val) setw(30) << setprecision(16) << chop(val) \
                               << setprecision(PREC)


// Triagonalisation by iteration.
//
// INPUT: tables du??_pos, du??_neg, de??_pos, de??_neg
// OUTPUT: written to files "xi??.dat" and "zeta??.dat"

// old2 == explicit -xi^2-zeta^2 terms, rather than contained in auxM,auxN
void tridiag_old2()
{
  ofstream XI11, ZETA11;
  safe_open(XI11, "xi11.dat"); // hopping constants 
  safe_open(ZETA11, "zeta11.dat"); // on-site energies
  ofstream XI12, ZETA12;
  safe_open(XI12, "xi12.dat"); // hopping constants 
  safe_open(ZETA12, "zeta12.dat"); // on-site energies
  ofstream XI21, ZETA21;
  safe_open(XI21, "xi21.dat"); // hopping constants 
  safe_open(ZETA21, "zeta21.dat"); // on-site energies
  ofstream XI22, ZETA22;
  safe_open(XI22, "xi22.dat"); // hopping constants 
  safe_open(ZETA22, "zeta22.dat"); // on-site energies
    
  mpf_set_default_prec(preccpp);
  cout << "Using precision of " << preccpp << " digits." << endl;

  // Constants
  mpf_class mpONE(1.e0);

  // Temporary MP variables
  Matrix temp, tempsq, sum, auxM, auxN;

  Matrix xiinv; // xi^(-1)
  Matrix mpxi; // xi
  Matrix xi2; // xi^2
  Matrix mpzeta; // zeta

  Matrix xi_prev, xi2_prev; // values in previous iteration

  xi_prev = zeroMAT;
  xi2_prev = zeroMAT;
   
  // Orthogonal transformation matrices
  vMatrix up(mMAX+1);
  vMatrix up_prev(mMAX+1);
  vMatrix up_prev2(mMAX+1);
  vMatrix um(mMAX+1);
  vMatrix um_prev(mMAX+1);
  vMatrix um_prev2(mMAX+1);

  // "Representative energy" matrices in the star representation.
  vMatrix ep1(mMAX+1);
  vMatrix em1(mMAX+1);

  // First index: site along the Wilson chain.
  // Second index: interval number in the star representation.
  mMatrix Up(Nmax+1, mMAX+1);
  mMatrix Um(Nmax+1, mMAX+1);
   
  cout << "Initializing." << endl;
   
  for (unsigned int m = 0; m <= mMAX; m++) {
    up_prev[m] = zeroMAT;
    up_prev[m](0,0) = du11_pos[m];
    up_prev[m](0,1) = du12_pos[m];
    up_prev[m](1,0) = du21_pos[m];
    up_prev[m](1,1) = du22_pos[m];
     
    um_prev[m] = zeroMAT;
    um_prev[m](0,0) = du11_neg[m];
    um_prev[m](0,1) = du12_neg[m];
    um_prev[m](1,0) = du21_neg[m];
    um_prev[m](1,1) = du22_neg[m];
     
    ep1[m] = zeroMAT;
    ep1[m](0,0) = de11_pos[m];
    ep1[m](0,1) = de12_pos[m];
    ep1[m](1,0) = de21_pos[m];
    ep1[m](1,1) = de22_pos[m];
    
    em1[m] = zeroMAT;
    em1[m](0,0) = -de11_neg[m]; // negative energies!
    em1[m](0,1) = -de12_neg[m];
    em1[m](1,0) = -de21_neg[m];
    em1[m](1,1) = -de22_neg[m];
     
    up_prev2[m] = zeroMAT;
    um_prev2[m] = zeroMAT;
  }

  fix_norm(up_prev, um_prev);

  // Store the first row of U
  for (unsigned int m = 0; m <= mMAX; m++) {
     Up(0, m) = up_prev[m];
     Um(0, m) = um_prev[m];
  }
   
  // Fix by Ziga, 8.1.2015 (was n <= Nmax before).
  for (unsigned int n = 0; n < Nmax; n++) {
    // Calculate zeta_n, xi2_n and xi_n

    mpzeta = zeroMAT;

    for (unsigned int m = 0; m <= mMAX; m++) {
      // up_prev = u^+_{n,m}
      tempsq = prod( ep1[m], trans(up_prev[m]) );
      mpzeta += prod( up_prev[m], tempsq );

      // um_prev = u^-_{n,m}
      tempsq = prod( em1[m], trans(um_prev[m]) );
      mpzeta += prod( um_prev[m], tempsq );
    }

    xi2 = zeroMAT;
    // auxiliary matrices
    for (unsigned int m = 0; m <= mMAX; m++) {
      auxM = prod( ep1[m], trans(up_prev[m]) );
      auxN = prod( em1[m], trans(um_prev[m]) );

      xi2 += prod( trans(auxM), auxM ) + prod( trans(auxN), auxN );
    }
    xi2 = xi2 - xi2_prev - prod(trans(mpzeta), mpzeta);

    if ( xi2(0,0) < 0.0 || xi2(1,1) < 0.0) {
      cerr << "xi2 negative, aborting." << endl;
      exit(1);
    }

    show("xi^2", xi2);
     
    mpxi = matsqrt(xi2);

    show("xi", mpxi);

    mpf_class d = matdet(mpxi);
    mpf_class xinorm2 = matnorm2(mpxi);

    cout << "det(xi)=" << d.get_d() << " norm2(xi)=" << xinorm2.get_d() << endl;
     
    if (abs(d.get_d()/xinorm2.get_d()) > 1e-12) { // invertible mpxi
       xiinv = matinv(mpxi);

       show("xi^-1", xiinv);

       // compute u_{n+1,m}, v_{n+1,m}
       for (unsigned int m = 0; m <= mMAX; m++) {
	  up[m] = prod( up_prev[m], trans(ep1[m]) ) - prod( trans(mpzeta), up_prev[m] ) - prod( trans(xi_prev), up_prev2[m] );
	  up[m] = prod( xiinv,  up[m] ); 
	  um[m] = prod( um_prev[m], trans(em1[m]) ) - prod( trans(mpzeta), um_prev[m] ) - prod( trans(xi_prev), um_prev2[m] );
	  um[m] = prod( xiinv, um[m] );
       }
    } else {
       // If det is small, this means xi has the structure of {{a sqrt(ab)},{sqrt(ab) b}}, because it is symmetric.
       // If rhs has the same structure, we may solve for up[m].
       cerr << "WARNING: determinant of xi is small!" << endl;

       // Special case: ((1,1),(1,1))
       if (is1111(mpxi)) {
	  cout << "xi propto ((1,1),(1,1))" << endl;
       
	  Matrix rhs;
       
	  // compute u_{n+1,m}, v_{n+1,m}
	  for (unsigned int m = 0; m <= mMAX; m++) {
	     rhs = prod( up_prev[m], trans(ep1[m]) ) - prod( trans(mpzeta), up_prev[m] ) - prod( trans(xi_prev), up_prev2[m] );
	     assert(is1111(rhs));
	     up[m] = oneMAT * (rhs(0,0)/mpxi(0,0));;
	     rhs = prod( um_prev[m], trans(em1[m]) ) - prod( trans(mpzeta), um_prev[m] ) - prod( trans(xi_prev), um_prev2[m] );
	     assert(is1111(rhs));
	     um[m] = oneMAT * (rhs(0,0)/mpxi(0,0));
	  }
       } else {
	  cerr << "to be implemented!" << endl;
	  exit(1);
       }
    }

    // After computing the new up and um, we fix their normalization. 
    fix_norm(up, um);

    // Store the next row of U
    for (unsigned int m = 0; m <= mMAX; m++) {
       Up(n+1, m) = up[m];
       Um(n+1, m) = um[m];
    }
    
    if (check_orthogonality) {
       for (unsigned int n1 = 0; n1 <= n+1; n1++) { 
	  cout << setw(3) << n1 << " U "; 
	  for (unsigned int n2 = n1; n2 <= n+1; n2++) { 
	     mpf_class overlap = 0.0; 
	     for (unsigned int m = 0; m <= mMAX; m++) { 
		const Matrix pdt1 = prod(Up(n1, m), trans(Up(n2, m))); 
		const Matrix pdt2 = prod(Um(n1, m), trans(Um(n2, m)));
		
		overlap += spur(pdt1) + spur(pdt2); 
	     } 
	     
	     cout << overlap.get_d() << " "; 
	  } 
	  cout << endl; 
       } 
    }

    // Save results
    {
       double dxi = mpxi(0,0).get_d();
       double coef_xi = dxi / SCALE(n+1);
       double dzeta = mpzeta(0,0).get_d();

       XI11 << coef_xi << endl;
       ZETA11 << dzeta << endl;

       cout << "  xi11(" << n << ")=" << HIGHPREC( dxi )
         << " --> " << HIGHPREC( coef_xi )<< endl;
       cout << "zeta11(" << n << ")=" << HIGHPREC( dzeta ) << endl;
    }
    {
       double dxi = mpxi(0,1).get_d();
       double coef_xi = dxi / SCALE(n+1);
       double dzeta = mpzeta(0,1).get_d();

       XI12 << coef_xi << endl;
       ZETA12 << dzeta << endl;

       cout << "  xi12(" << n << ")=" << HIGHPREC( dxi )
         << " --> " << HIGHPREC( coef_xi )<< endl;
       cout << "zeta12(" << n << ")=" << HIGHPREC( dzeta ) << endl;
    }
    {
       double dxi = mpxi(1,0).get_d();
       double coef_xi = dxi / SCALE(n+1);
       double dzeta = mpzeta(1,0).get_d();

       XI21 << coef_xi << endl;
       ZETA21 << dzeta << endl;

       cout << "  xi21(" << n << ")=" << HIGHPREC( dxi )
         << " --> " << HIGHPREC( coef_xi )<< endl;
       cout << "zeta21(" << n << ")=" << HIGHPREC( dzeta ) << endl;
    }
    {
       double dxi = mpxi(1,1).get_d();
       double coef_xi = dxi / SCALE(n+1);
       double dzeta = mpzeta(1,1).get_d();

       XI22 << coef_xi << endl;
       ZETA22 << dzeta << endl;

       cout << "  xi22(" << n << ")=" << HIGHPREC( dxi )
         << " --> " << HIGHPREC( coef_xi )<< endl;
       cout << "zeta22(" << n << ")=" << HIGHPREC( dzeta ) << endl;
    }
     
    // Store results from the previous iteration
    xi_prev = mpxi;
    xi2_prev = xi2;

    for (unsigned int m = 0; m <= mMAX; m++) {
      um_prev2[m] = um_prev[m];
      up_prev2[m] = up_prev[m];
      um_prev[m] = um[m];
      up_prev[m] = up[m];
    }
     
    cout << endl;
  }
}

int main(int argc, char *argv[])
{
   clock_t start_clock = clock();
   
   about();
   parser(param_fn);
   set_parameters();

   load_tables();
   tridiag_old2();
   
   clock_t end_clock = clock();
   cout << "# Elapsed " << double(end_clock-start_clock)/CLOCKS_PER_SEC << " s"
     << endl;
}
